/**
 * @file events.js
 * 
 * @brief This file contain implementation of trigger handling that executes
 *        comands on Sonos.
 * 
 * @author: mobile2DS
 * 
 * @date: 15.12.2013
 */
var lWarn = {
	logln : function(o) {
	}
};
var lInfo = {
	logln : function(o) {
	}
};

var LOGFILE_NAME = "sonos-controller.log";
lWarn = new Logger(LOGFILE_NAME);
lInfo = new Logger(LOGFILE_NAME);

/**
 * @brief Searches for Sonos group master Player
 * 
 * @param aGroups -
 *            array of all Sonos groups
 * @param sPlayerIP -
 *            ip of player
 * @param sDeviceID -
 *            unique id of player
 * 
 * @return if current player is slave, returns device ID of master player of
 *         group(coordinator), if current player is master result is empty
 * 
 */
function getPlayerGroupCoordinator(aGroups, sPlayerIP, sDeviceID) {
	var sCoordinator = "";
	if (aGroups != null) {
		for ( var iIndex = 0; iIndex < aGroups.length; iIndex++) {
			var sGroup = aGroups[iIndex].data;
			if (sGroup.indexOf(sPlayerIP) > -1) {
				var sMaster = sGroup.match(/Coordinator=\"(.*?)\"/);
				if (sMaster != null) {
					if (sMaster[1] != sDeviceID && sMaster[1].indexOf('RINCON') > -1) {
						sCoordinator = sMaster[1];
					}
				}
				break;
			}
		}
	}
	return sCoordinator;
}

function isSonosPlayerPluggedIn(aGroups, sPlayerIP) {
	if (aGroups != null) {
		for ( var iIndex = 0; iIndex < aGroups.length; iIndex++) {
			var sGroup = aGroups[iIndex].data;
			if (sGroup.indexOf(sPlayerIP) > -1) {
				return true;
			}
		}
	}
	return false;
}

/**
 * @brief Plays radio URL on Sonos player
 * 
 * @param sPlayerIP -
 *            ip of player where radio will be executed
 * @param sUrl -
 *            URL of radio station
 * @param sTitle -
 *            title of Radio station
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * 
 * @return
 */
function playRadio(sPlayerIP, sUrl, sTitle, curl_options) {
	upnpSetRadio(sPlayerIP, sUrl, sTitle, curl_options);
	upnpPlay(sPlayerIP, curl_options);
}

/**
 * @brief Executes Active-Queue of song stored in PT
 * 
 * @param sPlayerIP -
 *            ip of player where radio will be executed
 * @param sDeviceID -
 *            id of executed device
 * @param execAction -
 *            PT object of executed activity/preset
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * 
 * @return
 */
function playQueue(sPlayerIP, sDeviceID, execAction, curl_options) {
	var arrayQueue = execAction.getChild('active-queue').getChildren();
	if (arrayQueue != null) {
		lInfo.logln("Queue" + arrayQueue.length);
		// clear current queue
		upnpClearQueue(sPlayerIP, curl_options);
		for ( var yIndex = 0; yIndex < arrayQueue.length; yIndex++) {
			upnpLoadPlaylist(sPlayerIP, arrayQueue[yIndex].getChild('playID').getValue(), curl_options);
			if (yIndex == 0) { // start playing first song
				upnpSetContentSrc(sPlayerIP, "x-rincon-queue:" + sDeviceID + "#0", curl_options);
				upnpPlay(sPlayerIP, curl_options);
			}
			lInfo.logln("Adding " + arrayQueue[yIndex].getChild('playID').getValue());
		}
	}
}

/**
 * @brief Executes playPlaylist or single File on Sonos
 * 
 * @param sPlayerIP -
 *            ip of player where radio will be executed
 * @param sDeviceID -
 *            id of executed device
 * @param sFrom -
 *            item of execAction where read loaded comand
 *            (nowSingleUrl/playListURL)
 * @param execAction -
 *            PT object of executed activity/preset
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * @param sMediaGUID -
 *            In case of Media Lib file we send info of file ID, if necessary we
 *            fetch file url
 * 
 * @return
 */
function playPlaylist(sPlayerIP, sDeviceID, sFrom, execAction, curl_options, sMediaGUID) {
	upnpClearQueue(sPlayerIP, curl_options);
	var sLoadValue = execAction.getChild(sFrom).getValue();
	var metaData = null;
	lInfo.logln('play playlist ' + execAction.getPath());
	if (execAction.getChild('metaData') != null)
		metaData = unescape(execAction.getChild('metaData').getValue());
	if (sLoadValue != "") {
		upnpLoadPlaylist(sPlayerIP, sLoadValue, curl_options, metaData);

		// activating queue, in not active
		upnpSetContentSrc(sPlayerIP, "x-rincon-queue:" + sDeviceID + "#0", curl_options);
		upnpSeek(sPlayerIP, true, 1, curl_options);
		upnpPlay(sPlayerIP, curl_options);
	} else if (sMediaGUID != "" && execAction.getChild(sMediaGUID) != null) { // media
		// file
		// withouth
		// url
		var sGuidID = execAction.getChild(sMediaGUID).getValue();
		if (sGuidID != "") {
			cloudService.fetchMediaFile(unescape(sGuidID), function(oData) {
				if (oData.ReturnMessage == "OK") {
					var sUrl = unescape(oData.Response.UrlPath);
					storeSubParameter('sonos-activities', execAction.getChild("id").getValue() + '/nowSingleUrl', sUrl);

					upnpLoadPlaylist(sPlayerIP, sUrl, curl_options, metaData);

					// activating queue, in not active
					upnpSetContentSrc(sPlayerIP, "x-rincon-queue:" + sDeviceID + "#0", curl_options);
					upnpSeek(sPlayerIP, true, 1, curl_options);
					upnpPlay(sPlayerIP, curl_options);

					// nastavimo še dolžino komada
					objCurrentTrack = upnpGetCurrentTrack(sPlayerIP, curl_options);
					if (objCurrentTrack != null) {
						var nLength = secondToInt(unescape(objCurrentTrack.length));
						if (!isNaN(nLength))
							storeSubParameter('sonos-activities', execAction.getChild('id').getValue() + '/nowSingleLength', (nLength + 1));
					}
				}
			});
		}
	}
}

/**
 * @brief Set Sonos player volume
 * 
 * @param sPlayerIP -
 *            ip of player where radio will be executed
 * @param execAction -
 *            PT object of executed activity/preset
 * @param sGroupedSlavesIP -
 *            In case of Media Lib file we send info of file ID, if necessary we
 *            fetch file url
 * @param bSkipStop -
 *            if we skip stop command
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * 
 * @return
 */
function setPlayerVolume(sPlayerIP, execAction, sGroupedSlavesIP, bSkipStop, curl_options) {
	var bActionVolume = JSON.parse(execAction.getChild('volumeSet').getValue());
	if (bActionVolume != null && JSON.parse(bActionVolume)) {
		var nNewVolume = parseInt(execAction.getChild('volume').getValue());
		if (nNewVolume != null)
			if (!bSkipStop)
				upnpStop(sPlayerIP, curl_options);

		lInfo.logln("Exec command set volume " + nNewVolume);
		upnpSetVolume(sPlayerIP, nNewVolume, curl_options);

		// set volume also for all slave players in group
		// each player in group need to have assigned volume separately
		var aSlaves = null;
		if (sGroupedSlavesIP.length > 0) {
			aSlaves = sGroupedSlavesIP.split(",");
		}

		if (aSlaves != null && aSlaves.length > 0) {
			for ( var iIndex = 0; iIndex < aSlaves.length; iIndex++) {
				var sSlaveIP = aSlaves[iIndex].trim();
				lInfo.logln("Exec command set slave volume for " + sSlaveIP);
				upnpSetVolume(sSlaveIP, nNewVolume, curl_options);
			}
		}
	}
}

/**
 * @brief executes action on Sonos player
 * 
 * @param sPlayerIP -
 *            ip of player where radio will be executed
 * @param sDeviceID -
 *            id of executed device
 * @param execAction -
 *            PT object of executed activity/preset
 * @param sGroupedSlavesIP -
 *            In case of Media Lib file we send info of file ID, if necessary we
 *            fetch file url
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * 
 * @return
 */
function executeComandOnSonos(sPlayerIP, sDeviceID, execAction, sGroupedSlavesIP, curl_options) {
	var nActionType = parseInt(execAction.getChild('actionType').getValue());
	lInfo.logln("nActionType=" + nActionType);

	var bSetVolume = true;
	if (nActionType == SONOS_ACTION_COMMAND) {
		var nCommandID = parseInt(execAction.getChild('command').getValue());
		if (nCommandID == COMMAND_MUTE || nCommandID == COMMAND_UNMUTE)
			bSetVolume = false;
	}

	if (bSetVolume) {
		setPlayerVolume(sPlayerIP, execAction, sGroupedSlavesIP, false, curl_options);
	}

	switch (nActionType) {
	case SONOS_ACTION_NOW_PLAYING: // Now playing
		lInfo.logln("Now playing");
		if (execAction.getChild('nowAction').getValue() == 0) { // Single
			// track/Radio
			if (JSON.parse(execAction.getChild('nowSingleIsRadio').getValue())) { // radio
				lInfo.logln("Play radio " + execAction.getChild('nowSingleTitle').getValue());
				playRadio(sPlayerIP, execAction.getChild('nowSingleUrl').getValue(), execAction.getChild('nowSingleTitle').getValue(), curl_options);
			} else { // mp3
				lInfo.logln("Set " + execAction.getChild('playListURL').getValue());
				playPlaylist(sPlayerIP, sDeviceID, 'nowSingleUrl', execAction, curl_options, "");
			}
		} else { // Selected queue
			playQueue(sPlayerIP, sDeviceID, execAction, curl_options);
		}
		break;

	case SONOS_ACTION_PLAYLIST: // Playlist
		lInfo.logln("Playlist");
		playPlaylist(sPlayerIP, sDeviceID, 'playListURL', execAction, curl_options, "");
		break;

	case SONOS_ACTION_FAVORITES: // Sonos favorites
		lInfo.logln("Favorites play " + execAction.getChild('favoritesTitle').getValue());
		// radio favorite
		if (execAction.getChild('favoritesTitle').getValue().indexOf("TuneIn") > -1) {
			lInfo.logln("Favorites radio");
			playRadio(sPlayerIP, execAction.getChild('favoritesURL').getValue(), execAction.getChild('favoritesTitle').getValue(), curl_options);
		} else {
			lInfo.logln("Favorites other");
			// other favorites: replace queue, add to queue, play
			playPlaylist(sPlayerIP, sDeviceID, 'favoritesURL', execAction, curl_options, "");

			// if single mp3 is selected in favorites, we must fetch file lenght
			// (only if not defined yet).
			// It's not possible to get this info when saving activity in dialog
			if (execAction.getChild('nowSingleLength')!=null) {
				var sLength = execAction.getChild('nowSingleLength').getValue();
				if (sLength == "" || parseInt("" + sLength, 10) < 1) {
					objCurrentTrack = upnpGetCurrentTrack(sPlayerIP, curl_options);
					if (objCurrentTrack != null) {
						var nLength = secondToInt(unescape(objCurrentTrack.length));
						if (!isNaN(nLength))
							storeSubParameterFull(execAction.getPath() + '/nowSingleLength', (nLength + 1));// sRootPath
					}
				}
			}
		}
		break;

	case SONOS_ACTION_LIBRARY:
	case SONOS_ACTION_TTS:
		lInfo.logln("MediaLibrary/Text2Speech");

		var sGUID = "mediaFileGUID";
		if (nActionType == SONOS_ACTION_TTS)
			sGUID = "t2Sguid";

		lInfo.logln("playPlaylist " + sPlayerIP + "," + sDeviceID);
		playPlaylist(sPlayerIP, sDeviceID, 'nowSingleUrl', execAction, curl_options, sGUID);

		var sLength = execAction.getChild('nowSingleLength').getValue();
		if (sLength == "" || parseInt("" + sLength, 10) < 1) {
			objCurrentTrack = upnpGetCurrentTrack(sPlayerIP, curl_options);
			if (objCurrentTrack != null) {
				var nLength = secondToInt(unescape(objCurrentTrack.length));
				if (!isNaN(nLength))
					storeSubParameterFull(execAction.getPath() + '/nowSingleLength', (nLength + 1)); // sRootPath
			}
		}

		break;

	case SONOS_ACTION_COMMAND: // Command
		lInfo.logln("command=" + nCommandID);
		switch (nCommandID) {
		case COMMAND_STOP: // Stop
			upnpStop(sPlayerIP, curl_options);
			break;

		case COMMAND_PLAY: // Play
			upnpPlay(sPlayerIP, curl_options);
			break;

		case COMMAND_PAUSE: // Pause
			upnpPause(sPlayerIP, curl_options);
			break;

		case COMMAND_MUTE: // Mute
			upnpSetMute(sPlayerIP, true, curl_options);
			bSetVolume = false;
			break;

		case COMMAND_UNMUTE: // Unmute
			upnpSetMute(sPlayerIP, false, curl_options);
			break;
		}
		break;
	}
}

/**
 * @brief return list of players on which this activity must be executed
 * 
 * @param oActivity -
 *            PT object of executed activity
 * 
 * @return array of players with data object.id - player PT id object.device -
 *         player deviceID object.ip - player IP object.slaves - player list of
 *         slaves ID's object.IPSlaves - player list of slaves IP adresses
 */
function getPlayersToExecUndo(oActivity) {
	lInfo.logln('getPlayersToExecUndo');
	var arrayPlayers = new Array();
	if (oActivity != null) {
		var createSonosGroup = false;
		if (oActivity.getChild('createSonosGroup') != null)
			createSonosGroup = JSON.parse(oActivity.getChild('createSonosGroup').getValue());

		var baseNodeG = Property.getNode('sonos-groups');
		var oArrayNodeGroups = null;
		if (baseNodeG != null) {
			var oArrayNodeGroups = baseNodeG.getChildren();
			if (oArrayNodeGroups != null)
				lInfo.logln("oArrayNodeGroups = " + oArrayNodeGroups.length);
		}

		var nTmpSonosGroupID = 0;
		if (oActivity.getChild('tmpSonosGroup') != null)
			nTmpSonosGroupID = oActivity.getChild('tmpSonosGroup').getValue();

		if (JSON.parse(oActivity.getChild('allRooms').getValue())) {// all
			// players/rooms
			var baseNode = Property.getNode('sonos-players');
			if (baseNode != null) {
				var oArrayNode = baseNode.getChildren();
				for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
					if (!JSON.parse(oArrayNode[iIndex].getChild('disabled').getValue())) {
						if (!createSonosGroup || (createSonosGroup && isPlayerActiveInActivity(oArrayNodeGroups, nTmpSonosGroupID, oArrayNode[iIndex].getChild('id').getValue()))) {
							arrayPlayers.push(new Object({
								id : oArrayNode[iIndex].getChild('id').getValue(),
								device : oArrayNode[iIndex].getChild('deviceID').getValue(),
								ip : oArrayNode[iIndex].getChild('ipAddress').getValue(),
								slaves : oArrayNode[iIndex].getChild('playersListSlaves').getValue(),
								IPSlaves : oArrayNode[iIndex].getChild('playersListIPSlaves').getValue()
							}));
						}
					}
				}
			}
		} else {
			// only selected players/rooms
			if (oActivity.getChild('pickedRooms') != null) {
				var sRooms = oActivity.getChild('pickedRooms').getValue();
				if (sRooms != null) {
					var rooms = unescape(sRooms).split(";");
					if (rooms != null && rooms.length > 0) {
						for ( var iIndex = 0; iIndex < rooms.length; iIndex++)
							rooms[iIndex] = parseInt(rooms[iIndex], 10); // convert
						// to
						// int
						// to
						// avoid
						// indefOf
						// problem

						var oArrayNode = Property.getNode('sonos-players').getChildren();
						for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
							var nRoomID = oArrayNode[iIndex].getChild('roomID').getValue();

							if (!JSON.parse(oArrayNode[iIndex].getChild('disabled').getValue())) {
								if (rooms.indexOf(nRoomID) > -1) {
									if (!createSonosGroup || (createSonosGroup && isPlayerActiveInActivity(oArrayNodeGroups, nTmpSonosGroupID, oArrayNode[iIndex].getChild('id').getValue()))) {
										arrayPlayers.push(new Object({
											id : oArrayNode[iIndex].getChild('id').getValue(),
											device : oArrayNode[iIndex].getChild('deviceID').getValue(),
											ip : oArrayNode[iIndex].getChild('ipAddress').getValue(),
											slaves : oArrayNode[iIndex].getChild('playersListSlaves').getValue(),
											IPSlaves : oArrayNode[iIndex].getChild('playersListIPSlaves').getValue()
										}));
									}
								}
							}
						}
					}
				}
			}
		}
	}

	return arrayPlayers;
}

/**
 * @brief function reset temporary volume dimming steps
 * 
 * @return
 */
function resetDimming() {
	var dimmValue = Property.getNode("dimming");
	if (dimmValue != null) {
		var oArrayNode = dimmValue.getChildren();
		for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
			var oChildNode = oArrayNode[iIndex];
			dimmValue.removeChild(oChildNode);
		}
	}
}

/**
 * @brief testing if player is active in searched group
 * 
 * @param arrayGroups -
 *            list of stored active activities sonos groups
 * @param nGroupID -
 *            id of searched group
 * @param nPlayerID -
 *            id of player
 * 
 * @return true or false if player is active in this gruop
 */
function isPlayerActiveInActivity(arrayGroups, nGroupID, nPlayerID) {
	var bExecuteOnPlayer = false;
	if (arrayGroups != null && arrayGroups.length > 0) {
		for ( var z = 0; z < arrayGroups.length; z++) {
			var oChildGroup = arrayGroups[z];
			if (oChildGroup.getChild('id') != null && oChildGroup.getChild('id').getValue() == nGroupID) {
				var oGroupPlayers = oChildGroup.getChildren();
				for ( var x = 0; x < oGroupPlayers.length; x++) {
					var oChildNode = oGroupPlayers[x];
					if (oChildNode.getChild('playerID') != null && oChildNode.getChild('playerID').getValue() == nPlayerID) {
						bExecuteOnPlayer = true;
						break;
					}
				}
			}
		}
	}
	return bExecuteOnPlayer;
}

/**
 * @brief executes activity on Sonos, stores grouping info if needed, executes
 *        timers for timed activities
 * 
 * @param nSceneID -
 *            in case of nEventType=EVENT_TYPE_CALL_SCENE, we also need sceneID
 *            info
 * @param nEventType -
 *            type of activity EVENT_TYPE_CALL_SCENE/EVENT_TYPE_CUSTOM
 * @param sMasterPlayerIP -
 *            ip of master player
 * @param sMasterDevice -
 *            device ID of master player
 * @param bIsOneOfPlayersMuted -
 *            if one of player is muted we must to unmute it
 * @param arrayFilteredPlayers -
 *            array of players on which we will execute activity
 * @param iIndex -
 *            position of executed player inside arrayFilteredPlayers
 * @param execAction -
 *            PT object of executed activity
 * @param sendAction -
 *            object with settings that must be set on Sonos
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * 
 * @return
 */
function executePlayer(nSceneID, nEventType, sMasterPlayerIP, sMasterDevice, bIsOneOfPlayersMuted, arrayFilteredPlayers, iIndex, execAction, sendAction, curl_options) {
	var sPlayerIP = arrayFilteredPlayers[iIndex].ip;
	var sDeviceID = arrayFilteredPlayers[iIndex].device;
	if (sendAction.createSonosGroup) {
		sPlayerIP = sMasterPlayerIP;
		sDeviceID = sMasterDevice;
	} 
	
//	if (!sendAction.createSonosGroup) {
	// just remove any timer for that player if existing ... it will break the player behavior
	var playerID = arrayFilteredPlayers[iIndex].id;
	lInfo.logln('--  before remove TE ');
	if (Property.getNode("sonos-players/"+playerID+"/temporaryEventForReset")!=null) {	
		lInfo.logln('--  remove TE ' + Property.getProperty("sonos-players/"+playerID+"/temporaryEventForReset"));
		Property.getNode('/system/EventInterpreter/ScheduledEvents').removeChild(Property.getProperty("sonos-players/"+playerID+"/temporaryEventForReset"));
		Property.getNode("sonos-players/"+playerID).removeChild('temporaryEventForReset');
	}
//	}
	lInfo.logln('-- after remove TE ');


	lInfo.logln("Execute on player with IP " + sPlayerIP);
	lInfo.logln("sendAction=" + JSON.stringify(sendAction));

	if (sendAction.repeatLimited) {
		upnpSetPlayMode(sPlayerIP, "REPEAT_ALL", curl_options);
	} else if (arrayFilteredPlayers[iIndex].playMode != 'NORMAL')
		upnpSetPlayMode(sPlayerIP, "NORMAL", curl_options);

	//if (bIsOneOfPlayersMuted) { // if one player is muted we umute .... was maybe buggy!
		upnpSetMute(sPlayerIP, false, curl_options);
	//}

	var isSceneWithUndo = (predefinedUnduEvents.indexOf(nSceneID) > -1)
	executeComandOnSonos(sPlayerIP, sDeviceID, execAction, arrayFilteredPlayers[iIndex].IPSlaves, curl_options);

	// we set this value after execution command to sonos(only there we can get
	// track length)
	if (execAction.getChild('nowSingleLength')!=null)
		sendAction.nowSingleLength = parseInt("" + execAction.getChild('nowSingleLength').getValue(), 10);
	else
		sendAction.nowSingleLength = 0;
	var groupID = Math.floor(Math.random() * 100000000000);
	if (isSceneWithUndo) {
		// in case of undo scene we need to remember groupID, so we can get it
		// when undo trigger is executed
		storeSubParameter('sonos-activities', parseInt("" + execAction.getChild('id').getValue(), 10) + '/tmpSonosGroup', groupID);
	} else
		storeSubParameter('sonos-activities', parseInt("" + execAction.getChild('id').getValue(), 10) + '/tmpSonosGroup', 0);

	if (arrayFilteredPlayers != null && arrayFilteredPlayers.length > 0) {
		if (sendAction.createSonosGroup)
			storeSubParameter('sonos-groups', groupID + '/id', groupID);

		// remove player from other grouping activities, and if activity is
		// grouped we assign active players
		for ( var y = 0; y < arrayFilteredPlayers.length; y++) {
			var playerID = arrayFilteredPlayers[y].id;
			var baseNode = Property.getNode('sonos-groups');
			if (baseNode != null) {
				var oArrayNode = baseNode.getChildren();
				for ( var z = 0; z < oArrayNode.length; z++) {
					var oChildGroup = oArrayNode[z];
					var oGroupPlayers = oChildGroup.getChildren();
					for ( var x = 0; x < oGroupPlayers.length; x++) {
						var oChildNode = oGroupPlayers[x];
						if (oChildNode.getChild('playerID') != null && oChildNode.getChild('playerID').getValue() == playerID) {
							oChildGroup.removeChild(oChildNode.getChild('id').getValue());
						}
					}
				}
			}

			if (sendAction.createSonosGroup) {
				storeSubParameter('sonos-groups', groupID + '/' + y + '/id', y);
				storeSubParameter('sonos-groups', groupID + '/' + y + '/playerID', playerID);
			}
		}
	}

	if (sendAction.repeatLimited || (sendAction.createSonosGroup && !isSceneWithUndo) || (sendAction.restartAfter && !isSceneWithUndo)) { // set
		// timer
		// if
		// time
		// limited
		// execution,
		// or
		// not
		// undo
		// event
		// with
		// restart
		// data
		var execLength = sendAction.repeatTime;

		if (nSceneID != -1) { // if not testing button
			if (sendAction.nowSingleLength > 0 && !sendAction.repeatLimited) { // defined
				// exec
				// time
				// by
				// mp3
				// file
				// time
				lInfo.logln("Set new Length=" + sendAction.nowSingleLength);
				execLength = sendAction.nowSingleLength;
			}
		}

		// in case of length problem we set predefined default exec time
		if (execLength == 0)
			execLength = DEF_REPEAT_TIME;

		if (!sendAction.createSonosGroup) { // no player grouping, create event
			// for each speaker
			var timerData = new Object();
			timerData.sceneID = nSceneID;
			timerData.id = arrayFilteredPlayers[iIndex].id;
			timerData.eventType = nEventType;
			timerData.playerIP = arrayFilteredPlayers[iIndex].ip;
			timerData.deviceID = sMasterDevice;
			timerData.onlyStop = (sendAction.repeatLimited && !sendAction.restartAfter);

			lInfo.logln("New timer will execeute after =" + execLength);
			
			var TimedEventID=(new TimedEvent('sonos_execute_timer', '+' + execLength,{
				path : JSON.stringify(timerData)
			})).raise();
			Property.setProperty("sonos-players/"+arrayFilteredPlayers[iIndex].id + "/temporaryEventForReset",TimedEventID);
		} else {
			// players grouping create single event
			/*var timerData = new Object();
			timerData.sceneID = nSceneID;
			timerData.eventType = nEventType;
			timerData.playerMasters = arrayFilteredPlayers;
			timerData.onlyStop = (sendAction.repeatLimited && !sendAction.restartAfter);
			timerData.groupID = groupID; */

			lInfo.logln("New group timer will execeute after =" + execLength);

			/*var TimedEventID=(new TimedEvent('sonos_execute_timer_group','+' + execLength, {
				path : JSON.stringify(timerData)
			})).raise();
			Property.setProperty(execAction.getPath() + "/temporaryEventForReset",TimedEventID);*/

			if (Property.getNode("groups/temporaryEventForReset")!=null) {
				Property.getNode('/system/EventInterpreter/ScheduledEvents').removeChild(Property.getProperty("groups/temporaryEventForReset"));
				Property.getNode("groups").removeChild('temporaryEventForReset')
			}

			// only a timer when the stack should be new popped
			var TimedEventID=(new TimedEvent('sonos_execute_timer_group','+' + execLength, {
				//path : JSON.stringify(timerData)
			})).raise();
			Property.setProperty("groups/temporaryEventForReset",TimedEventID);
			
		}
	}
}

/**
 * @brief Executions Stored Sonos data
 * 
 * @param arrayGroups -
 *            list of stored active activities sonos groups
 * @param sDeviceID -
 *            id of executed device
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * 
 * @return
 */
function executeRestoredData(pSonosData, deviceID, curl_options) {
	if (pSonosData != null) {
		lInfo.logln("Executing executeRestoredData");

		upnpClearQueue(pSonosData.playerIP, curl_options);
		if (pSonosData.activeQueueID != "") {
			// loadingOldQueue
			var aTmp = pSonosData.activeQueueID.split(':');
			if (aTmp != null && aTmp.length == 2) {
				var sQueueNr = aTmp[1];
				upnpLoadPlaylist(pSonosData.playerIP, "file:///jffs/settings/savedqueues.rsq#" + sQueueNr, curl_options);
			}
			lInfo.logln("Deleting tmp queue");
			upnpDeleteQueueFromList(pSonosData.playerIP, pSonosData.activeQueueID, curl_options);
		}

		if (pSonosData.isMuted) {
			upnpSetMute(pSonosData.playerIP, true, curl_options);
		}
		if (pSonosData.volumeReset) {
			lInfo.logln("Reset volume to " + pSonosData.volumeDef);
			upnpSetVolume(pSonosData.playerIP, pSonosData.volumeDef, curl_options);
		}
		if (pSonosData.playMode != "") {
			upnpSetPlayMode(pSonosData.playerIP, pSonosData.playMode, curl_options);
		}

		if (!pSonosData.isRadio) {
			lInfo.logln("Restarting music:" + "x-rincon-queue:" + deviceID + "#0" + "/" + pSonosData.activeAction);
			// activating queue
			upnpSetContentSrc(pSonosData.playerIP, "x-rincon-queue:" + deviceID + "#0", curl_options);
			upnpSeek(pSonosData.playerIP, true, pSonosData.trackListPos, curl_options);
			upnpSeek(pSonosData.playerIP, false, pSonosData.trackPlayPos, curl_options);
			if (pSonosData.activeAction == "PLAYING")
				upnpPlay(pSonosData.playerIP, curl_options);
		} else {
			lInfo.logln("Restarting radio:" + pSonosData.playUrl + "/" + pSonosData.activeAction);
			upnpStop(pSonosData.playerIP, curl_options);
			upnpSetRadio(pSonosData.playerIP, pSonosData.playUrl, "Radio", curl_options);
			if (pSonosData.activeAction == "PLAYING")
				upnpPlay(pSonosData.playerIP, curl_options);
		}
	}
}

function storeOldGroupingSettings() {
	var oPlayerArray=getAllPlayers();
	var arraySonosGroups = upnpGetSonosGrouping(oPlayerArray, curl_options);
	lInfo.logln("active groups: " + JSON.stringify(arraySonosGroups));
	if (Property.getNode("groups/dSGroup")!=null) {
		var myGroups=Property.getNode("groups/dSGroup").getChildren();
		for (var iIndex=0;iIndex<myGroups.length;iIndex++) {
			var myGroup=myGroups[iIndex].getValue();
			for (var jIndex=0;jIndex<arraySonosGroups.length;jIndex++)
				if (arraySonosGroups[jIndex].id==myGroup) {
					lInfo.logln('found a ds group in the groups, so NO storage should be executed!')
					return;
				}
		}
		// no ds groups are present, so remove obsolete data
		Property.getNode('groups').removeChild("dSGroup");		
	}
	lInfo.logln('no dSGroup are in action, so we store it');
	// getting old volumne information

	for (var iIndex=0;iIndex<arraySonosGroups.length;iIndex++) {
		for (var jIndex=0;jIndex<arraySonosGroups[iIndex].members.length;jIndex++) {
			arraySonosGroups[iIndex].members[jIndex].volume=upnpGetVolume(arraySonosGroups[iIndex].members[jIndex].ip,curl_options);
			arraySonosGroups[iIndex].members[jIndex].mute=upnpGetMute(arraySonosGroups[iIndex].members[jIndex].ip, curl_options);
		}
	}
	
	var baseNode = Property.getNode('sonos-players');
	if (baseNode == null) 
		return;
	var oArrayNode = baseNode.getChildren();
	for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) 
	if (("" + oArrayNode[iIndex].getChild('notDiscovered').getValue()) != "true") {
		
		sPlayerIP= oArrayNode[iIndex].getChild('ipAddress').getValue();
		var restartDataForPlayer = {};
		restartDataForPlayer.playerIP=sPlayerIP;
		restartDataForPlayer.id=oArrayNode[iIndex].getChild('deviceID').getValue();
		restartDataForPlayer.isRadio = true;
		restartDataForPlayer.playMode = upnpGetPlayMode(sPlayerIP, curl_options);

		restartDataForPlayer.trackListPos = "";
		restartDataForPlayer.trackPlayPos = "";
		restartDataForPlayer.isMuted=false; // done externally, so only placeholder for executeRestoredData
		restartDataForPlayer.volumeReset=false; // done externally, so only placeholder for executeRestoredData
		restartDataForPlayer.activeAction = upnpGetTransportInfo(sPlayerIP, curl_options);
		restartDataForPlayer.activeQueueID = upnpSaveQueue(sPlayerIP, restartDataForPlayer.id, curl_options);
			
		var sMedia = upnpGetMediaInfo(sPlayerIP, curl_options);
		if (sMedia != null && sMedia != "") {
			matches = sMedia.match(/<CurrentURI>(.*?)<\/CurrentURI>/);
			var nrTracks = 0;
			if (matches != null) {
				restartDataForPlayer.playUrl = matches[1];
				if (matches[1].indexOf("x-rincon-queue") > -1) { // active queue
					restartDataForPlayer.isRadio = false;
					matches = sMedia.match(/<NrTracks>(.*?)<\/NrTracks>/);
					if (matches != null)
						nrTracks = parseInt("" + matches[1], 10);

					lInfo.logln("nrTracks=" + nrTracks);
					if (nrTracks > 0) { // remembering current track nr., music pos.
						restartDataForPlayer.trackListPos = 1;
						restartDataForPlayer.trackPlayPos = "0:00:00";

						var sPosInfo = upnpGetPositionInfo(sPlayerIP, curl_options);
						if (sPosInfo != null && sPosInfo != "") {
							matches = sPosInfo.match(/<Track>(.*?)<\/Track>/);
							if (matches != null)
								restartDataForPlayer.trackListPos = parseInt("" + matches[1], 10);
							matches = sPosInfo.match(/<RelTime>(.*?)<\/RelTime>/);
							if (matches != null)
								restartDataForPlayer.trackPlayPos = matches[1];
						}
						lInfo.logln("trackListPos=" + restartDataForPlayer.trackListPos);
						lInfo.logln("trackPlayPos=" + restartDataForPlayer.trackPlayPos);
					}
				}
			}
		}
		Property.setSavedProperty(oArrayNode[iIndex].getPath()+'/resumeDataAfterGrouping',JSON.stringify(restartDataForPlayer));
	}		
	Property.setSavedProperty("groups/usergrouping",JSON.stringify(arraySonosGroups));
}

function resetToOldGrouping(ignorePlayerID) {
	lInfo.logln('Starting resetToOldGrouping');
	if (Property.getNode("groups/dSGroup")== null)
		return;
	lInfo.logln('weiter mit resetToOldGrouping');
	
	if (Property.getNode("groups/temporaryEventForReset")!=null) {
		Property.getNode('/system/EventInterpreter/ScheduledEvents').removeChild(Property.getProperty("groups/temporaryEventForReset"));
		Property.getNode("groups").removeChild('temporaryEventForReset')
	}
	
	var oldGroups=JSON.parse(Property.getProperty("groups/usergrouping"));
	
	for ( var iIndex = 0; iIndex < oldGroups.length; iIndex++) {
		var oOldGroup=oldGroups[iIndex];
		var sMasterPlayerIP=null;
		var sMasterDevice=oOldGroup.master;

		for ( var jIndex = 0; jIndex < oOldGroup.members.length; jIndex++) {
			var oDevice=oOldGroup.members[jIndex];
			if (oDevice.id==sMasterDevice)
				sMasterPlayerIP=oDevice.ip;
		}		
		
		upnpSetPlayerAsStandAloneGroup(sMasterPlayerIP, curl_options);
		for ( var jIndex = 0; jIndex < oOldGroup.members.length; jIndex++) {
			var oDevice=oOldGroup.members[jIndex];
			if (oDevice.id!=sMasterDevice)
				upnpSetContentSrc(oDevice.ip, "x-rincon:" + sMasterDevice, curl_options);
			upnpSetVolume(oDevice.ip, oDevice.volume, curl_options);
		}		
	}	
	Property.getNode('groups').removeChild("dSGroup");
	Property.getNode('groups').removeChild("dSGroupStack");
	
	lInfo.logln('resetToOldGrouping: resuming music before grouping');
	
	var baseNode = Property.getNode('sonos-players');
	if (baseNode == null) 
		return;
	var oArrayNode = baseNode.getChildren();
	for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) 
	if (("" + oArrayNode[iIndex].getChild('notDiscovered').getValue()) != "true") {
		if ((ignorePlayerID==-1) || (ignorePlayerID!=oArrayNode[iIndex].getChild('id').getValue())) 
			if (oArrayNode[iIndex].getChild('resumeDataAfterGrouping')!=null) {
				var oRestoreData=JSON.parse(oArrayNode[iIndex].getChild('resumeDataAfterGrouping').getValue());
				var deviceID=oArrayNode[iIndex].getChild('deviceID').getValue();
				executeRestoredData(oRestoreData,deviceID, curl_options);
				oArrayNode[iIndex].removeChild('resumeDataAfterGrouping');
			}
			
	}
}

function makeNewGrouing(players,desiredVolumne) {
	storeOldGroupingSettings();
	var sMasterDevice = players[0].device;
	var sMasterPlayerIP = players[0].ip;

	upnpStop(sMasterPlayerIP, curl_options);
	upnpSetPlayerAsStandAloneGroup(sMasterPlayerIP, curl_options);
	if (desiredVolumne!=-1)
		upnpSetVolume(sMasterPlayerIP, desiredVolumne, curl_options);
	
	groupID = upnpGetSonosGroupID(sMasterPlayerIP, sMasterDevice, curl_options);
	lInfo.logln('makeNewGrouing: -- Master [' + sMasterPlayerIP + '] now OUR new Group is ' + groupID);
	Property.setSavedProperty('groups/dSGroup/'+ groupID,groupID);
	
	for ( var iIndex = 1; iIndex < players.length; iIndex++) {

		// connect sonos to temporary group
		lInfo.logln("makeNewGrouing: Assign player[" + players[iIndex].ip + "] to group " + sMasterDevice);
	
		upnpStop(players[iIndex].ip, curl_options);

		// add player to group
		upnpSetContentSrc(players[iIndex].ip, "x-rincon:" + sMasterDevice, curl_options);
		// must set volume for each slave parent separately 
		if (desiredVolumne!=-1)
			upnpSetVolume(players[iIndex].ip, desiredVolumne, curl_options);
	}
	for ( var iIndex = 0; iIndex < players.length; iIndex++) {
		// check if had slaves, add slaves to group
		if (players[iIndex].slaves!="") {
			var arraySlaves = players[iIndex].IPSlaves.split(';');
			for (var jIndex=0;jIndex<arraySlaves.length;jIndex++) {
				var sSlaveIP = arraySlaves[jIndex].trim();			
				lInfo.logln("Also adding slave "+sSlaveIP);
		        upnpSetContentSrc(sSlaveIP, "x-rincon:"+sMasterDevice, curl_options);
		    	if (desiredVolumne!=-1)
		    		upnpSetVolume(sSlaveIP, desiredVolumne, curl_options);
			}
		}
	}
}

function popGroupStack() {
	var pStackNode=Property.getNode('groups/dSGroupStack');
	if (pStackNode!=null) {
		var stackNodes=pStackNode.getChildren();
		lInfo.logln('Stack has ' + stackNodes.length);
		if (stackNodes.length<2) {
			// end of stack
			resetToOldGrouping(-1);
			return -1;
		}
		
		var lastNode=stackNodes[stackNodes.length-1];
		var newLastNode=stackNodes[stackNodes.length-2];
		pStackNode.removeChild(lastNode.getName());
		return newLastNode.getChild('activityID').getValue();
	} else {
		resetToOldGrouping(-1);
		return -1;
	}
	
}

function pushGroupStack(pActionNode,length) {
	if (pActionNode==null)
		return;
	var myID=parseInt(''+pActionNode.getChild('id').getValue(),10);
	var pStackNode=Property.getNode('groups/dSGroupStack');
	var pIndex=0;
	if (pStackNode!=null) {
		var stackNodes=pStackNode.getChildren();
		for (var iIndex=0;iIndex<stackNodes.length;iIndex++) {
			if (parseInt(stackNodes[iIndex].getName(),10)>=pIndex)
				pIndex=parseInt(stackNodes[iIndex].getName(),10)+1;
			if (stackNodes[iIndex].getChild('activityID').getValue()==myID)
				Property.getNode('groups/dSGroupStack').removeChild(stackNodes[iIndex].getName());
		}
	}
	Property.setSavedProperty('groups/dSGroupStack/' + pIndex + '/activityID',myID);	
}

/**
 * @brief Executes triggered activity
 * 
 * @param sActivityPath -
 *            PT path of activity
 * @param nSceneID -
 *            in case of nEventType=EVENT_TYPE_CALL_SCENE, we also need sceneID
 *            info
 * @param nEventType -
 *            type of activity EVENT_TYPE_CALL_SCENE/EVENT_TYPE_CUSTOM
 * @param bExecuteAsTrigger -
 *            true executed from trigger, false executed after dSS restart
 * @param curl_options -
 *            object with info of connectionTimeout for executed upnp command
 * 
 * @return
 */
function executeActivity(sActivityPath, nSceneID, nEventType, bExecuteAsTrigger, curl_options) {
	var execAction = Property.getNode(sActivityPath);
	if (execAction != null) {
		var sendAction = null;
		var listPlayers = new Array();
		var execLength=0;
		lInfo.logln(" ------------- Prepareing activity  ------------------")
//		lInfo.logln("Scene recieved=" + nSceneID);

//		lInfo.logln("Scene final= " + nSceneID + ", type " + nEventType);
		var bSpecialScenesOnlyRestore = false; // for sceneID > 64

		sendAction = new Object();
		sendAction.volumeSet = JSON.parse(execAction.getChild('volumeSet').getValue());
		sendAction.volume = execAction.getChild('volume').getValue();
		sendAction.actionType = execAction.getChild('actionType').getValue();
		sendAction.restartAfter = JSON.parse(execAction.getChild('restartAfter').getValue());
		if ((!sendAction.restartAfter) && (nSceneID > 64) && (sendAction.actionType != SONOS_ACTION_COMMAND) && (nEventType == EVENT_TYPE_CALL_SCENE)) {
			sendAction.restartAfter = true;
			bSpecialScenesOnlyRestore = true;
		}

		sendAction.repeatLimited = JSON.parse(execAction.getChild('setRepeat').getValue());
		sendAction.createSonosGroup = false;
		if (execAction.getChild('createSonosGroup') != null)
			sendAction.createSonosGroup = JSON.parse(execAction.getChild('createSonosGroup').getValue());

		sendAction.repeatTime = 0;
		// repeat for maximum of x minutes, converted to seconds
		if (sendAction.repeatLimited) {
			sendAction.repeatTime = parseInt("" + execAction.getChild('repeatTime').getValue(), 10) * 60;
			execLength=sendAction.repeatTime;
		} else {
			if (execAction.getChild('nowSingleLength')!=null)
				execLength=execAction.getChild('nowSingleLength').getValue();
			else 
				execLength=0;
		}
		sendAction.playInAllRooms = JSON.parse(execAction.getChild('allRooms').getValue());
		sendAction.filteredRooms = new Array();
		if (!sendAction.playInAllRooms)
			sendAction.filteredRooms = unescape(execAction.getChild('pickedRooms').getValue()).split(';');

		if (nSceneID == -1) {// test button called from UI - Main Grid!!!!!
			if (sendAction.createSonosGroup)
				sendAction.restartAfter = true; // allways restart if grouping
			// is checked

			sendAction.repeatTime = TEST_PLAY_TIME; // test time before restore

			var sLength = "";
			if (execAction.getChild('nowSingleLength')!=null) 
				sLength=execAction.getChild('nowSingleLength').getValue();
			if (sLength != "" && parseInt("" + sLength, 10) > 0 && parseInt("" + sLength, 10) < TEST_PLAY_TIME) {
				sendAction.repeatTime = sLength;
				execLength=sendAction.repeatTime;
			}
		}

		lInfo.logln('-- base information known: ' + JSON.stringify(sendAction));
		if (sendAction != null) {
			var arrayFilteredPlayers = new Array();

			// 2. list of players to send command
			var baseNode = Property.getNode('sonos-players');
			if (baseNode != null) {
				var arrayPlayers = baseNode.getChildren();
				if (arrayPlayers != null) {

					// reading all sonoses groups only once
					var arraySonosGroups = upnpGetSonosGroups(getAllPlayers(), curl_options);
					//lInfo.logln("arraySonosGroups=" + JSON.stringify(arraySonosGroups));

					for ( var iIndex = 0; iIndex < arrayPlayers.length; iIndex++) {
						var playerData = arrayPlayers[iIndex];
						if (!JSON.parse(playerData.getChild('disabled').getValue())) {
							// for each player which is not disabled

							var playerRoomID = playerData.getChild('roomID').getValue();
							var sPlayerIP = playerData.getChild('ipAddress').getValue();
							var playerID = playerData.getChild('id').getValue();
							var sDeviceID = playerData.getChild('deviceID').getValue();
							var sIPSlaves = playerData.getChild('playersListIPSlaves').getValue();

							// skipping rooms that are not included in activity
							if (sendAction.playInAllRooms || (!sendAction.playInAllRooms && sendAction.filteredRooms.indexOf(playerRoomID.toString()) > -1)) {

								// testing if player is currently plugged in
								if (isSonosPlayerPluggedIn(arraySonosGroups, sPlayerIP)) {

									if (!sendAction.createSonosGroup && !isActivityGroupingActive(playerID)) {
										var sCoordinator = getPlayerGroupCoordinator(arraySonosGroups, sPlayerIP, sDeviceID);

										// current player is in group and not
										// master, so we skip him
										if (sCoordinator.length > 0) {
											// room player is part of group and
											// not coordinator, we set sPlayerIP
											// to coordinator of this group
											for ( var xIndex = 0; xIndex < arrayPlayers.length; xIndex++) {
												if (arrayPlayers[xIndex].getChild('deviceID').getValue() == sCoordinator) {
													sPlayerIP = arrayPlayers[xIndex].getChild('ipAddress').getValue();
													sDeviceID = sCoordinator;
													break;
												}
											}
										}
									}
									// if been triggered and not restored after
									// a startup .... should be normal
									if (bExecuteAsTrigger) {
										var handleResult = handleNewActivity(playerID, sendAction.createSonosGroup, nSceneID, nEventType, COMMAND_PLAY, sendAction.restartAfter, execAction.getChild(
												'id').getValue(), 0);
										lInfo.logln("handleResult[" + playerData.getChild('ipAddress').getValue() + "]=" + JSON.stringify(handleResult));
										// activity is not postponed because of
										// a blinking action ... should be
										// normal!
										if (!handleResult.storeActivity) {
											var bStateOk = true;
											var sPlayerPlayMode = "";
											var bPlayerMute = false;
											if (handleResult.storeCurrDevData) {
												bPlayerMute = upnpGetMute(sPlayerIP, curl_options);
											}
											if (sendAction.restartAfter && handleResult.storeCurrDevData) {
												var restartData = new Object();
												restartData.id = playerData.getChild('id').getValue();
												restartData.isRadio = true;
												restartData.volumeDef = "";
												restartData.playMode = "";
												restartData.trackListPos = "";
												restartData.trackPlayPos = "";
												restartData.sceneID = nSceneID;
												restartData.playerIP = sPlayerIP;
												restartData.volumeReset = true;
												restartData.activeAction = "STOPPED";
												restartData.isMuted = bPlayerMute;
												if (!bSpecialScenesOnlyRestore)
													restartData.activeAction = upnpGetTransportInfo(sPlayerIP, curl_options);
												restartData.activeQueueID = upnpSaveQueue(sPlayerIP, playerID, curl_options);
												lInfo.logln('-- fetching old informations about player '+sPlayerIP);
												var sMedia = upnpGetMediaInfo(sPlayerIP, curl_options);
												if (sMedia != null && sMedia != "") {
													matches = sMedia.match(/<CurrentURI>(.*?)<\/CurrentURI>/);
													var nrTracks = 0;
													if (matches != null) {
														restartData.playUrl = matches[1];
														if (matches[1].indexOf("x-rincon-queue") > -1) { // active
															// queue
															restartData.isRadio = false;
															matches = sMedia.match(/<NrTracks>(.*?)<\/NrTracks>/);
															if (matches != null)
																nrTracks = parseInt("" + matches[1], 10);

															lInfo.logln("nrTracks=" + nrTracks);
															if (nrTracks > 0) { // remembering
																// current
																// track
																// nr.,
																// music
																// pos.
																restartData.trackListPos = 1;
																restartData.trackPlayPos = "0:00:00";

																var sPosInfo = upnpGetPositionInfo(sPlayerIP, curl_options);
																if (sPosInfo != null && sPosInfo != "") {
																	matches = sPosInfo.match(/<Track>(.*?)<\/Track>/);
																	if (matches != null)
																		restartData.trackListPos = parseInt("" + matches[1], 10);
																	matches = sPosInfo.match(/<RelTime>(.*?)<\/RelTime>/);
																	if (matches != null)
																		restartData.trackPlayPos = matches[1];
																}
																lInfo.logln("trackListPos=" + restartData.trackListPos);
																lInfo.logln("trackPlayPos=" + restartData.trackPlayPos);
															}
														}
													}

													restartData.volumeDef = upnpGetVolume(sPlayerIP, curl_options);
													// store slaves volume
													// values
													var slavesVolumes = "";
													if (sIPSlaves.length > 0) {
														aSlaves = sIPSlaves.split(",");

														if (aSlaves != null && aSlaves.length > 0) {
															for ( var xIndex = 0; xIndex < aSlaves.length; xIndex++) {
																var sSlaveIP = aSlaves[xIndex].trim();
																var nTmpVolume = upnpGetVolume(sSlaveIP, curl_options);
																if (slavesVolumes.length > 0)
																	slavesVolumes = slavesVolumes + ",";
																if (nTmpVolume != null)
																	slavesVolumes = slavesVolumes + nTmpVolume
																else
																	slavesVolumes = slavesVolumes + "-1"; // error
																// by
																// reading
																// value
															}
														}
													}
													restartData.volumeSlaves = slavesVolumes;
													sPlayerPlayMode = upnpGetPlayMode(sPlayerIP, curl_options);
													restartData.playMode = sPlayerPlayMode;

													lInfo.logln('-- Restart Data: ' + JSON.stringify(restartData));
													bStateOk = storeUpnpDeviceStatus(restartData, sendAction.createSonosGroup);
												} // failed while fetching
												// player mediaInfo
												else {
													bStateOk = false;
												}
											} else {
												lInfo.logln("-- no restart data, no restart should be executed");
											}

											if (handleResult.executeUpnpCmd && bStateOk)
												arrayFilteredPlayers.push(new Object({
													"id" : playerID,
													"ip" : sPlayerIP,
													"device" : sDeviceID,
													"slaves" : playerData.getChild('playersListSlaves').getValue(),
													"IPSlaves" : sIPSlaves,
													"playMode" : sPlayerPlayMode,
													"isMuted" : bPlayerMute,
													"restoreData" : null,
												}));
										} else {
											lInfo.logln("Store activity ... but why ?");
											storeActivityForLater(playerID, sendAction.createSonosGroup, execAction.getChild('id').getValue());

											if (handleResult.executeUpnpCmd && handleResult.restoreData != null) {
												arrayFilteredPlayers.push(new Object({
													"id" : playerID,
													"ip" : sPlayerIP,
													"device" : sDeviceID,
													"slaves" : playerData.getChild('playersListSlaves').getValue(),
													"IPSlaves" : sIPSlaves,
													"playMode" : handleResult.playMode,
													"isMuted" : handleResult.isMuted,
													"restoreData" : handleResult
												}));
											}
										}
									} else {
										// executing after dssRestart
										bPlayerMute = upnpGetMute(sPlayerIP, curl_options);
										sPlayerPlayMode = upnpGetPlayMode(sPlayerIP, curl_options);

										arrayFilteredPlayers.push(new Object({
											"id" : playerID,
											"ip" : sPlayerIP,
											"device" : sDeviceID,
											"slaves" : playerData.getChild('playersListSlaves').getValue(),
											"IPSlaves" : sIPSlaves,
											"playMode" : sPlayerPlayMode,
											"isMuted" : bPlayerMute,
											"restoreData" : null
										}));
									}
								}
							}
						} else {
							lInfo.logln("Player(" + playerData.getChild('sonosRoom').getValue() + ") skipped while not selected as active");
						}
					}
					lInfo.logln("-- now all preps are done, executing things for " + JSON.stringify(arrayFilteredPlayers));
					if (arrayFilteredPlayers.length > 0) {
						var sMasterDevice = "";
						var sMasterPlayerIP = "";

						var bMasterStopped = false;
						var bIsOneOfPlayersMuted = false;
						var groupID = null;
						
						for ( var iIndex = 0; iIndex < arrayFilteredPlayers.length; iIndex++) {
							if (arrayFilteredPlayers[iIndex].isMuted)
								bIsOneOfPlayersMuted = true;
						}
						var fShouldCreateGroup=sendAction.createSonosGroup;
						if (arrayFilteredPlayers.length<2) // play on only 1 player-> no group needed
							fShouldCreateGroup=false;
						if (fShouldCreateGroup) {
							var newVolume=-1;	
							var bActionVolume = JSON.parse(execAction.getChild('volumeSet').getValue());
							if (bActionVolume != null && JSON.parse(bActionVolume)) {
								newVolume = parseInt(execAction.getChild('volume').getValue());
							}
							
							makeNewGrouing(arrayFilteredPlayers,newVolume);
							pushGroupStack(execAction,execLength);
							if (arrayFilteredPlayers.restoreData == null)
								executePlayer(nSceneID, nEventType, arrayFilteredPlayers[0].ip, arrayFilteredPlayers[0].device, bIsOneOfPlayersMuted, arrayFilteredPlayers, 0, execAction, sendAction, curl_options);
							else
								executeRestoredData(arrayFilteredPlayers.restoreData, arrayFilteredPlayers[0].device, curl_options);
						} else {
							lInfo.logln('-- removing the Grouping');
							resetToOldGrouping(-1); // ensure that grouping is removed
							for ( var iIndex = 0; iIndex < arrayFilteredPlayers.length; iIndex++) {
								lInfo.logln('make player ' + JSON.stringify(arrayFilteredPlayers[iIndex]));								
								if (arrayFilteredPlayers[iIndex].restoreData == null)
									executePlayer(nSceneID, nEventType, arrayFilteredPlayers[iIndex].ip, arrayFilteredPlayers[iIndex].device, bIsOneOfPlayersMuted, arrayFilteredPlayers, iIndex, execAction, sendAction, curl_options);
								else
									executeRestoredData(arrayFilteredPlayers[iIndex].restoreData, arrayFilteredPlayers[iIndex].device, curl_options);
							}
						}
						

						if (nSceneID == -1 && !sendAction.restartAfter) {
							var evt = new Event("sonos-test-activity-finished", {});
							evt.raise();
							lInfo.logln('Test activity finished, no restore of previous sonos players states is needed.');
						}
					} else {
						if (nSceneID == -1) {
							var evt = new Event("sonos-test-activity-finished", {
								activityPath : sActivityPath
							});
							evt.raise();
							lInfo.logln('Failed to execute activity');
						}
					}
				}
			}
		}

		lInfo.logln("Activity START property.store");
		Property.store();
		lInfo.logln("Activity FINISHED property.store");
	}
}

/**
 * @brief Executes triggered Preset for room
 * 
 * @param sPresetPath -
 *            PT path of executed Preset
 * 
 * @return
 */
function executePreset(eventParam) {
	var execAction = Property.getNode(eventParam.path);
	if (execAction != null) {
		var nZoneID = parseInt(execAction.getChild('roomID').getValue());
		var baseNode = Property.getNode('sonos-players');
		var arrayPlayers = new Array();
		if (baseNode != null) {
			var arrayPlayers = baseNode.getChildren();
			var sPlayerIP = "";
			var sDeviceID = "";
			var sGroupedSlavesIP = "";
			if (arrayPlayers != null) {
				for ( var iIndex = 0; iIndex < arrayPlayers.length; iIndex++) {
					if (arrayPlayers[iIndex].getChild('roomID').getValue() == nZoneID) {
						if (!JSON.parse(arrayPlayers[iIndex].getChild('disabled').getValue())) {
							sPlayerIP = arrayPlayers[iIndex].getChild('ipAddress').getValue();
							lInfo.logln('-- removing the Grouping for ID ' + arrayPlayers[iIndex].getChild('id').getValue());
							resetToOldGrouping(arrayPlayers[iIndex].getChild('id').getValue());
							sDeviceID = arrayPlayers[iIndex].getChild('deviceID').getValue();
							if (arrayPlayers[iIndex].getChild('playersListIPSlaves') != null)
								sGroupedSlavesIP = arrayPlayers[iIndex].getChild('playersListIPSlaves').getValue();
							
							
							var playerID = arrayPlayers[iIndex].getChild('id').getValue();
							lInfo.logln('-- removing TE for ID ' + playerID);;
							if (Property.getNode("sonos-players/"+playerID+"/temporaryEventForReset")!=null) {	
								lInfo.logln('--  remove TE ' + Property.getProperty("sonos-players/"+playerID+"/temporaryEventForReset"));
								Property.getNode('/system/EventInterpreter/ScheduledEvents').removeChild(Property.getProperty("sonos-players/"+playerID+"/temporaryEventForReset"));
								Property.getNode("sonos-players/"+playerID).removeChild('temporaryEventForReset');
							}
							
							break;
						}
					}
				}
			}
			
			
			
			lInfo.logln("zone[" + nZoneID + "], assigned sPlayerIP=" + sPlayerIP + "/id=" + sDeviceID);
			// test if player moved to other group, then assign to master room
			var arraySonosGroups = upnpGetSonosGroups(getAllPlayers(), curl_options);
			
			// test if player is active/plugged in
			var bIsPlayerActive = false;
			if (arraySonosGroups != null && arraySonosGroups.length > 0) {
				bIsPlayerActive = isSonosPlayerPluggedIn(arraySonosGroups, sPlayerIP);
			}

			if (bIsPlayerActive) {
				var sCoordinator = getPlayerGroupCoordinator(arraySonosGroups, sPlayerIP, sDeviceID);
				lInfo.logln("sCoordinator=" + sCoordinator);
				// var sCoordinator = upnpGetGroupCoordinator(sPlayerIP,
				// sDeviceID, curl_options);
				if (sCoordinator.length > 0) {
					// room player is part of group and not coordinator, we set
					// sPlayerIP to coordinator of this group
					for ( var iIndex = 0; iIndex < arrayPlayers.length; iIndex++) {
						if (arrayPlayers[iIndex].getChild('deviceID').getValue() == sCoordinator) {
							sPlayerIP = arrayPlayers[iIndex].getChild('ipAddress').getValue();
							sDeviceID = sCoordinator;
							break;
						}
					}
				}
				lInfo.logln("action execution sPlayerIP=" + sPlayerIP);
				if (sPlayerIP != "") {
					if (upnpGetMute(sPlayerIP, curl_options))
						upnpSetMute(sPlayerIP, false, curl_options);

					executeComandOnSonos(sPlayerIP, sDeviceID, execAction, sGroupedSlavesIP, curl_options);
					upnpSetPlayMode(sPlayerIP, "NORMAL", curl_options);
				}
			} else {
				if (eventParam.testPreset != null && eventParam.testPreset == 1) {
					var evt = new Event('sonos-test-preset-executed', {
						presetPath : eventParam.path
					});
					evt.raise();
					lInfo.logln('Returning error for testing not available Preset');
				}
			}
		}
	}
}

var curl_options = new Object({
	'debug' : 0,
	'connectTimeout' : 200
});
lInfo.logln("-----------------------------");
lInfo.logln(JSON.stringify(raisedEvent));

if (raisedEvent.name === 'sonos_execute' || raisedEvent.name === 'sonos_execute_pause') {
	var start = new Date().getTime();
	try {
		if (raisedEvent.parameter != null) {
			var nGroupID = -1;
			if (raisedEvent.parameter.groupID != null)
				nGroupID = parseInt("" + raisedEvent.parameter.groupID, 10);
			lInfo.logln("nGroupID  " + nGroupID);
			lInfo.logln("path  " + raisedEvent.parameter.path);

			// reseting dimming tmp values
			resetDimming();

			if (nGroupID == 4 || raisedEvent.name === 'sonos_execute_pause') { // check
				// and
				// execute
				// audio
				// actions
				executePreset(raisedEvent.parameter);
			} else {
				// execute triggered activities

				// default as costum event handling
				var nSceneID = 0;
				var nEventType = EVENT_TYPE_CUSTOM;

				if (typeof raisedEvent.parameter.sceneID !== 'undefined') {
					nSceneID = parseInt(raisedEvent.parameter.sceneID);
					nEventType = EVENT_TYPE_CALL_SCENE;
				} else {
					if (typeof raisedEvent.parameter.statename !== 'undefined') {
						nSceneID = raisedEvent.parameter.statename;
						if (typeof raisedEvent.parameter.state !== 'undefined')
							nSceneID += raisedEvent.parameter.state;

						nEventType = EVENT_TYPE_CALL_SCENE;
						// undefined sceneID's
						switch (nSceneID) {
						case 'fireactive':
							nSceneID = 76;
							break;
						case 'rainactive':
							nSceneID = 88;
							break;
						case 'raininactive':
							nSceneID = 89;
							break;
						case 'hailactive':
							nSceneID = 90;
							break;
						case 'hailinactive':
							nSceneID = 91;
							break;
						case 'windactive':
							nSceneID = 86;
							break;
						case 'windinactive':
							nSceneID = 87;
							break;
						default:
							nEventType = EVENT_TYPE_CUSTOM;
							nSceneID = 0;
							break;
						}
					}
				}

				executeActivity(raisedEvent.parameter.path, nSceneID, nEventType, true, curl_options);
			}
		}
		lInfo.logln("END sonos_execute");
	} catch (e) {
		lInfo.logln("Error by sonos_execute: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}
} else if (raisedEvent.name === 'sonos_execute_undo') {
	try {
		lInfo.logln("undoScene zoneID/groupID/sceneID:  " + raisedEvent.source.zoneID + " / " + raisedEvent.source.groupID + " / " + raisedEvent.parameter.sceneID);

		var nGroupID = parseInt("" + raisedEvent.source.groupID, 10);
		var nSceneID = parseInt("" + raisedEvent.parameter.sceneID, 10);

		if (typeof raisedEvent.parameter.sceneID === 'undefined') {
			switch (raisedEvent.parameter.statename) {
			case 'fire':
				nSceneID = 76;
				break;
			}
		}

		var sMainPath = raisedEvent.parameter.path.replace("/undoCallScene", "");
		lInfo.logln("Path " + sMainPath);
		var execAction = Property.getNode(sMainPath);

		if (execAction != null) {// && predefinedUnduEvents.indexOf(nSceneID)
			// > -1)
			var restartData = null;
			var nTmpGroupID = 0;
			if (execAction.getChild('tmpSonosGroup') != null)
				nTmpGroupID = parseInt("" + execAction.getChild('tmpSonosGroup').getValue(), 10);

			// find sceneID in activities
			var bGrouping = JSON.parse(execAction.getChild('createSonosGroup').getValue());
			if (bGrouping) {
				popGroupStack();
			} else {
				var arrayPlayers = getPlayersToExecUndo(execAction);
	
				lInfo.logln("UNDO ExecPlayers:" + JSON.stringify(arrayPlayers));
	
				if (arrayPlayers != null && arrayPlayers.length > 0) {
					var arraySonosGroups = null;
					for ( var iIndex = 0; iIndex < arrayPlayers.length; iIndex++) {
						var sPlayerIP = arrayPlayers[iIndex].ip;
						var sDeviceID = arrayPlayers[iIndex].device;
						var sPlayerID = arrayPlayers[iIndex].id;
	
/*						if (bGrouping) {
							if (arraySonosGroups == null)
								arraySonosGroups = upnpGetSonosGroups(getAllPlayers(), curl_options);
	
							upnpSetPlayerAsStandAloneGroup(sPlayerIP, curl_options);
							addSlavePlayersToGroup(arrayPlayers[iIndex].slaves, arraySonosGroups, sDeviceID, curl_options);
						}
*/	
						var undoData = handleUndoActivity(sPlayerID, bGrouping, nSceneID);
						if (undoData != null) {
							var restartData = undoData.restoreData;
							lInfo.logln("Undo restart data:" + JSON.stringify(restartData));
							if (restartData != null) {
								executeRestoredData(restartData, sDeviceID, curl_options);
							} else {
								if (undoData.performStop) {
									lInfo.logln("No sonos player undo, stop event music");
									upnpStop(sPlayerIP, curl_options);
								}
							}
						}
					}
	
/*					if (bGrouping) {
						var oGroups = Property.getNode('sonos-groups');
						if (oGroups != null) {
							oGroups.removeChild(nTmpGroupID);
						}
					}*/
					lInfo.logln("sonos_execute_undo START property.store");
					Property.store();
					lInfo.logln("sonos_execute_undo START property.store");
				}
			}
		}

		lInfo.logln("end undo scene");
	} catch (e) {
		lInfo.logln("Error by undoScene: " + e);
		for ( var i in e) {
			lInfo.logln(i + ': ' + e[i]);
		}
	}

} else if (raisedEvent.name === 'sonos_execute_timer') {
	var timerData = JSON.parse(unescape(raisedEvent.parameter.path));
	// lInfo.logln(JSON.stringify(timerData));

	if (timerData != null) {
		// reseting dimming values
		resetDimming();

		var playerInfo = handleActivityTimeout(timerData.id, 0, timerData.sceneID, timerData.eventType);

		lInfo.logln("after exec handleActivityTimeout");
		if (playerInfo.restoreData != null) {
			if (timerData.onlyStop) {
				upnpStop(timerData.playerIP, curl_options);
			} else {
				lInfo.logln("handleActivityTimeout return data");
				executeRestoredData(playerInfo.restoreData, timerData.deviceID, curl_options);
			}
		} else if (playerInfo.activityId != -1) {
			lInfo.logln("handleActivityTimeout return activityID=" + playerInfo.activityId)
			var execAction = Property.getNode('/scripts/sonos-controller/sonos-activities/' + playerInfo.activityId);
			executeComandOnSonos(timerData.playerIP, timerData.deviceID, execAction, "", curl_options);
			upnpSetPlayMode(timerData.playerIP, "REPEAT_ALL", curl_options);
		} else if (playerInfo.performStop) {
			lInfo.logln("No data, stop event music");
			upnpStop(timerData.playerIP, curl_options);
		}

		lInfo.logln("sonos_execute_timer START property.store");
		Property.store();
		lInfo.logln("sonos_execute_timer START property.store");
		if (timerData.sceneID == -1) {
			var evt = new Event("sonos-test-activity-finished", {});
			evt.raise();
		}
	}
} else if (raisedEvent.name === 'sonos_execute_timer_group') {
	lInfo.logln("Grouped speakers timer finishing");
	resetDimming();
	var restoreToID=popGroupStack();
	if (restoreToID!=-1) {
		var nSceneID = 0;     // WHO THE FUCK CARES! JUST PLAY THAT THINGIE
		var nEventType = EVENT_TYPE_CUSTOM;

		executeActivity('sonos-activities/' + restoreToID, nSceneID, nEventType, true, curl_options);
	}
	var evt = new Event("sonos-test-activity-finished", {});
	evt.raise();
	lInfo.logln("sonos_execute_timer_group START property.store");
	Property.store();
	lInfo.logln("sonos_execute_timer_group START property.store");
}
